//==============================================================================
// Selection tutorial
// 2002-2004 by Sascha Willems
// http://www.delphigl.de
//==============================================================================
// Updated 19.01.2004 :
//  - Renamed TForm1 to TGLForm
//  - Replaced OpenGL12.pas with dglOpenGL.pas (GL1.5)
//  - DC and RC now released on destroy
//==============================================================================
unit selection1;

interface

uses
  Windows, Messages, SysUtils, Variants, Classes, Graphics, Controls, Forms,
  Dialogs, dglOpenGL, AppEvnts, GLBmp, ExtCtrls, StdCtrls;

type
  TGLForm = class(TForm)
    ApplicationEvents1: TApplicationEvents;
    Timer1: TTimer;
    Panel1: TPanel;
    Timer2: TTimer;
    Label1: TLabel;
    Label2: TLabel;
    Label3: TLabel;
    procedure FormCreate(Sender: TObject);
    procedure ApplicationEvents1Idle(Sender: TObject; var Done: Boolean);
    procedure Timer1Timer(Sender: TObject);
    procedure FormResize(Sender: TObject);
    procedure Timer2Timer(Sender: TObject);
    procedure FormClick(Sender: TObject);
    procedure FormMouseMove(Sender: TObject; Shift: TShiftState; X,
      Y: Integer);
    procedure FormDestroy(Sender: TObject);
  private
  public
    FDC    : HDC;
    FRC    : HGLRC;
    List   : Cardinal;
    yrot   : TGLFloat;
    Frames : Integer;
    mx,my  : Word;
    function GetSelectBufferHit : Integer;
    procedure GLInit;
    procedure LoadTextures;
    procedure DrawScene;
  end;

var
  Texture     : array[0..4] of TGLBmp;
  GLForm      : TGLForm;
  DisplayList : array[0..3] of Cardinal;
  Rotation    : array[0..3] of TGLFloat;
  SelfRot     : array[0..3] of TGLFloat;

const
 // These are names for our planets to use for selection
 Sun   = 0;
 Mars  = 1;
 Earth = 2;
 Moon  = 3;

implementation

{$R *.dfm}

// =============================================================================
// Generates a Sphere and returns the number of the displaylist
// =============================================================================
function CreateSphere(CX, CY, CZ, Radius : TGLFloat; N : Integer) : Cardinal;
var
 i,j                  : Integer;
 theta1,theta2,theta3 : TGLFloat;
 X, Y, Z, px, py, pz  : TGLFloat;
begin
Result := glGenLists(1);
glNewList(Result, GL_COMPILE);
if Radius < 0 then
 Radius :=-Radius;
if n < 0 then
 n := -n;
if (n < 4) or (Radius <= 0) then
 begin
 glBegin(GL_POINTS);
  glVertex3f(CX, CY, CZ);
 glEnd;
 exit;
 end;
for j := 0 to n div 2 -1 do
 begin
 theta1 := J*2*PI/N - PI/2;
 theta2 := (J+1)*2*PI/n - PI/2;
 glBegin(GL_QUAD_STRIP);
 for i := 0 to n do
  begin
  theta3 := i*2*PI/N;
  x := cos(theta2) * cos(theta3);
  y := sin(theta2);
  z := cos(theta2) * sin(theta3);
  px := CX + Radius*x;
  py := CY + Radius*y;
  pz := CZ + Radius*z;
  glNormal3f(X,Y,Z);
  glTexCoord2f(1-I/n, 2*(J+1)/n);
  glVertex3f(px,py,pz);
  X := cos(theta1) * cos(theta3);
  Y := sin(theta1);
  Z := cos(theta1) * sin(theta3);
  px := CX + Radius*X;
  py := CY + Radius*Y;
  pz := CZ + Radius*Z;
  glNormal3f(X,Y,Z);
  glTexCoord2f(1-i/n, 2*j/n);
  glVertex3f(px,py,pz);
  end;
 glEnd;
 end;
glEndList;
end;

procedure TGLForm.FormCreate(Sender: TObject);
begin
InitOpenGL;
// Get a handle from our Window
FDC := GetDC(Handle);
// Generate our rendering context with a color depth of 32Bpp
FRC := CreateRenderingContext(FDC, [opDoubleBuffered], 32, 24, 0, 0, 0, 0);
ActivateRenderingContext(FDC, FRC);
Frames := 0;
GLInit;
LoadTextures;
end;

// =============================================================================
// Sets up all needed OpenGL properties
// =============================================================================
procedure TGLForm.GLInit;
var
 nStackDepth : TGLUInt;
begin
glEnable(GL_TEXTURE_2D);
glEnable(GL_DEPTH_TEST);
glEnable(GL_LIGHTING);
glEnable(GL_COLOR_MATERIAL);
glLightModeli(GL_LIGHT_MODEL_LOCAL_VIEWER, GL_TRUE);
DisplayList[0] := CreateSphere(0,0,0,48,100); // Sun
DisplayList[1] := CreateSphere(0,0,0,16,100); // Mars
DisplayList[2] := CreateSphere(0,0,0,20,100); // Earth
DisplayList[3] := CreateSphere(0,0,0,8,75);   // Moon
glGetIntegerV(GL_MAX_NAME_STACK_DEPTH, @nStackDepth);
Label3.Caption := 'Depth of Name Stack : '+IntToStr(nStackDepth);
end;

// =============================================================================
// Loads all needed textures
// =============================================================================
procedure TGLForm.LoadTextures;
const
 Name : array[0..4] of String = ('sun','mars','earth','moon','space');
var
 i : Integer;
begin
for i := 0 to High(Name) do
 begin
 Texture[i] := TGLBmp.Create;
 Texture[i].LoadImage(Name[i]+'.jpg');
 Texture[i].GenTexture(False,False);
 end;
end;

// =============================================================================
// Draws the Scene
// =============================================================================
procedure TGLForm.DrawScene;
const
 Position   : array[0..3] of TGLFLoat = (0, 0, 0, 0);
 Ambient    : array[0..3] of TGLFloat = (0.3, 0.3, 0.3, 0);
 Diffuse    : array[0..3] of TGLFloat = (0.8, 0.8, 0.8, 0);
 Specular   : array[0..3] of TGLFloat = (0.8, 0.8, 0.8, 0);
 mSpecular  : array[0..3] of TGLFloat = (1.0, 1.0, 1.0, 1.0);
 mShininess : TGLFloat = 10.0;
begin
glMatrixMode(GL_MODELVIEW);
glClear(GL_DEPTH_BUFFER_BIT or GL_COLOR_BUFFER_BIT);
glLoadIdentity;
glMaterialfv(GL_FRONT, GL_SPECULAR, @mSpecular);
glMaterialfv(GL_FRONT, GL_SHININESS, @mShininess);
// Enable one Light and set it's properties
glEnable(GL_LIGHTING);
glEnable(GL_LIGHT0);
glLightfv(GL_LIGHT0, GL_POSITION, @Position);
glLightfv(GL_LIGHT0, GL_AMBIENT, @Ambient);
glLightfv(GL_LIGHT0, GL_DIFFUSE, @Diffuse);
glLightfv(GL_LIGHT0, GL_SPECULAR, @specular);
glTranslatef(0,0,-500);
glInitNames;
glPushName(0);
// Sun
glLoadName(Sun);       // Push the ID of the sun on the namestack
Texture[0].Bind;
glPushMatrix;
glRotatef(SelfRot[0], 0, 1, 0);
glCallList(DisplayList[0]);
glPopMatrix;
// Mars
glLoadName(Mars);       // Push the ID of the mars on the namestack
Texture[1].Bind;
glPushMatrix;
glRotatef(Rotation[1], 0, 1, 0);
glTranslatef(0,0,100);
glRotatef(SelfRot[1], 0, 1, 0);
glCallList(DisplayList[1]);
glPopMatrix;
// Earth
glLoadName(Earth);       // Push the ID of the earth on the namestack
Texture[2].Bind;
glPushMatrix;
glRotatef(Rotation[2], 0, 1, 0);
glTranslatef(0,0,200);
glRotatef(SelfRot[2], 0, 1, 0);
glCallList(DisplayList[2]);
// Moon
glLoadName(Moon);       // Push the ID of the moon on the namestack
Texture[3].Bind;
glRotatef(Rotation[3], 0, 1, 0);
glTranslatef(0,0,35);
glCallList(DisplayList[3]);
glPopMatrix;
// Background
// Note that you need to load another name on the stack.Otherwise
// the background would have the same name attached to as the moon!
glLoadName(1000);
Texture[4].Bind;
glBegin(GL_QUADS);
 glTexCoord2f(0, 0); glVertex3f(-470,  -350, -300);
 glTexCoord2f(0, 2); glVertex3f(-470,   350, -300);
 glTexCoord2f(2, 2); glVertex3f( 470,   350, -300);
 glTexCoord2f(2, 0); glVertex3f( 470,  -350, -300);
glEnd;
SwapBuffers(FDC);
end;

procedure TGLForm.ApplicationEvents1Idle(Sender: TObject;
  var Done: Boolean);
begin
DrawScene;
Done := False;
inc(Frames);
end;

procedure TGLForm.Timer1Timer(Sender: TObject);
var
 i : Integer;
begin
SelfRot[0]  := SelfRot[0]+0.1;  // Sun
SelfRot[1]  := SelfRot[1]+1;    // Mars
SelfRot[2]  := SelfRot[2]+1;    // Earth
SelfRot[3]  := SelfRot[3]+1;    // Moon
for i := 0 to High(SelfRot) do
 if SelfRot[i] > 360 then
  SelfRot[i] := 0;
Rotation[3] := Rotation[3]-2.5; // Moon
Rotation[2] := Rotation[2]+0.5; // Earth
Rotation[1] := Rotation[1]+0.3; // Mars
for i := 0 to High(Rotation) do
 begin
 if Rotation[i] > 360 then
  Rotation[i] := 0;
 if Rotation[i] < 0 then
  Rotation[i] := 360;
 end;
end;

procedure TGLForm.FormResize(Sender: TObject);
begin
if HandleAllocated then
 begin
 glViewport(0, 0, ClientWidth, ClientHeight);
 glMatrixMode(GL_PROJECTION);
 glLoadIdentity;
 gluPerspective(45, ClientWidth/ClientHeight, 0.1, 1000);
 end;
end;

procedure TGLForm.Timer2Timer(Sender: TObject);
begin
Caption := 'www.DelphiGL.de -Tutorial no.3 - Selection - '+IntToStr(Frames)+' FPS';
Frames := 0;
end;

function TGLForm.GetSelectBufferHit : Integer;
var
 SelectBuffer : array[0..512] of TGLUInt;
 Viewport     : TGLArrayi4;
 Hits,i       : Integer;
 HitZValue    : TGLUInt;
 Hit          : TGLUInt;
begin
// Store our Viewport
glGetIntegerv(GL_VIEWPORT, @viewport);
// Tell OpenGL the size of our select buffer
glSelectBuffer(512, @SelectBuffer);
// Change our render mode to GL_SELECT
glRenderMode(GL_SELECT);
// Init the namestack
glInitNames;
glPushName(0);
glMatrixMode(GL_PROJECTION);
glPushMatrix;
glLoadIdentity;
// Zoom the view to one pixel around the mouse position
gluPickMatrix(mx, viewport[3]-my, 1.0, 1.0, viewport);
gluPerspective(45.0, ClientWidth/ClientHeight, 0.1, 1000);
// Now we draw our scene in the select render mode
DrawScene;
glMatrixMode(GL_PROJECTION);
glPopMatrix;
// Now store the number of hits in Hits
Hits := glRenderMode(GL_RENDER);
// Now we look which hit has the lowest Z-Value i.e. was in front
Hit       := High(TGLUInt);
HitZValue := High(TGLUInt);
for i := 0 to Hits-1 do
 if SelectBuffer[(i*4)+1] < HitZValue then
  begin
  Hit       := SelectBuffer[(i*4)+3];
  HitZValue := SelectBuffer[(i*4)+1];
  end;
Result := Hit;
end;

procedure TGLForm.FormClick(Sender: TObject);
var
 Hit : Integer;
begin
Hit := GetSelectBufferHit;
case Hit of
 0 : Label2.Caption := 'sun';
 1 : Label2.Caption := 'mars';
 2 : Label2.Caption := 'earth';
 3 : Label2.Caption := 'moon';
else
 Label2.Caption := 'none';
end;
end;

procedure TGLForm.FormMouseMove(Sender: TObject; Shift: TShiftState; X,
  Y: Integer);
begin
mx := x;
my := y;
end;

procedure TGLForm.FormDestroy(Sender: TObject);
begin
DeactivateRenderingContext;
wglDeleteContext(FRC);
ReleaseDC(Handle, FDC);
end;

end.
